/*!
 *  @file GEENYmodem.cpp
 *
 *  @mainpage Arduino library for the GEENYmodem GPRS modem.
 *
 *  @section intro_sec Introduction
 *
 * 	Arduino library for the GEENYmodem GPRS modem.
 *
 *  @section dependencies Dependencies
 *  This library depends on the tinyGSM library, the HTTPClient library, and the PubSub library.
 
 *
 *  @section author Author
 *
 *  René Bohne
 *
 * 	@section license License
 *
 * 	MIT (see license.txt)
 */

#include "Arduino.h"
#include "GEENYmodem.h"

/**
 * @brief Constructor for GEENYmodem objects
 * 
 *
 */
GEENYmodem::GEENYmodem(void)
{
  gprs = NULL;
  client = NULL;
  mqtt = NULL;
  debug = NULL;
}

/**
 * @brief Destructor
 *
 */
GEENYmodem::~GEENYmodem(void)
{
  if (gprs)
  {
    delete gprs;
  }
  if (client)
  {
    delete client;
  }
  if (mqtt)
  {
    delete mqtt;
  }
}

/*!
 *    @brief  Sets up the connection to the GEENYmodem hardware.
 *    @param  port
 *            The serial port that connects to the modem
 *    @return True if connection to modem was successful.
 */
bool GEENYmodem::begin(Stream &port)
{
  if (gprs)
  {
    delete gprs;
  }
  if (client)
  {
    delete client;
  }
  if (mqtt)
  {
    delete mqtt;
  }

  gprs = new TinyGsm(port);
  gprs->restart();

  client = new TinyGsmClient(*gprs);
  mqtt = new PubSubClient(*client);

  return true;
}

/*!
 *    @brief  Sets up the connection to the GEENYmodem hardware.
 *    @param  port
 *            The serial port that connects to the modem
 *    @param  debugport
 *            The serial port that connects to the debug console
 *    @return True if connection to modem was successful.
 */
bool GEENYmodem::begin(Stream &port, Stream &debugport)
{
  if (gprs)
  {
    delete gprs;
  }
  if (client)
  {
    delete client;
  }
  if (mqtt)
  {
    delete mqtt;
  }

  debug = &debugport;
  debug->println("geenymodem v1.0");

  gprs = new TinyGsm(port);
  gprs->restart();

  client = new TinyGsmClient(*gprs);
  mqtt = new PubSubClient(*client);

  return true;
}

/**
 * @brief Get the Modem Information
 *
 * @return modem info String
 */
String GEENYmodem::getModemInfo(void)
{
  return gprs->getModemInfo();
}

/**
 * @brief Get the Signal Quality
 *
 * @return signal quality as unsigned 16bit int
 */
uint16_t GEENYmodem::getRSSI(void)
{
  return gprs->getSignalQuality();
}

/**
 * @brief Get the Network Connection Status
 *
 * @return true if network is connected
 */
bool GEENYmodem::isNetworkConnected(void)
{
  debug->println("GEENYmodem::isNetworkConnected");
  return gprs->isNetworkConnected();
}

/**
 * @brief Gets the CCID of a sim card via AT+CCID
 *
 * @return CCID as String
 */
String GEENYmodem::getSimCCID(void)
{
  return gprs->getSimCCID();
}

/**
 * @brief Asks for TA Serial Number Identification (IMEI) of SIM800C
 *
 * @return IMEI as String
 */

String GEENYmodem::getIMEI(void)
{
  return gprs->getIMEI();
}

/**
 * @brief Asks for International Mobile Subscriber Identity IMSI of SIM card
 *
 * @return IMSI as String
 */

String GEENYmodem::getIMSI(void)
{
  return gprs->getIMSI();
}

/**
 * @brief Asks for operator name
 *
 * @return operator name as String
 */
String GEENYmodem::getOperator(void)
{
  return gprs->getOperator();
}

/**
 * @brief get IP address
 *
 * @return IP address as String
 */
String GEENYmodem::getLocalIP(void)
{
  return gprs->getLocalIP();
}

void GEENYmodem::mqttCallback(char *topic, byte *payload, unsigned int len)
{
  //currently, we don't support mqtt subscriptions, but this is how a typical callback would look like:
  /*
  debug->print("Message arrived [");
  debug->print(topic);
  debug->print("]: ");
  debug->write(payload, len);
  debug->println();
  */

  // Only proceed if incoming message's topic matches
  /*
  if (String(topic) == topicLed) {
    ledStatus = !ledStatus;
    digitalWrite(LED_PIN, ledStatus);
    mqtt->publish(topicLedStatus, ledStatus ? "1" : "0");
  }
  */
}

/**
 * @brief Get the MQTT Connection Status
 *
 * @return true if network is connected
 */
bool GEENYmodem::isMQTTConnected(void)
{
  return this->mqttstatus;
}

/*!
 *    @brief  Disconnects from MQTT broker
 */
void GEENYmodem::disconnectMQTT(void)
{
  this->mqttstatus = false;
  mqtt->disconnect();
}

/*!
 *    @brief  Connects to MQTT broker
 *    @return True if connection to MQTT broker was successful.
 */
boolean GEENYmodem::connectMQTT(void)
{
  debug->println("GEENYmodem::connectMQTT");
  if (!this->isNetworkConnected())
  {
    if (!this->connectGPRS())
    {
      debug->println("NO GPRS CONNECTION");
      this->mqttstatus = false;
      return false;
    }
  }
  if (!poweredup)
  {
    if (!this->powerup()) //get mqtt credentials from modemmanager
    {
      debug->println("NOT REGISTERED WITH MODEM MANAGER");
      this->mqttstatus = false;
      return false;
    }
  }

  //id, user, password
  debug->print("thing_id: ");
  debug->println(this->thing_id);
  debug->print("thing_key: ");
  debug->println(this->thing_key);
  mqtt->setServer(MQTT_BROKER_URL, MQTT_BROKER_PORT);
  if (!mqtt->connect(this->thing_id, "thing", this->thing_key))
  {
    debug->println("MQTT CONNECTION WRONG CREDENTIALS");
    this->mqttstatus = false;
    return false;
  }
  debug->println("MQTT CONNECTION OK");
  this->mqttstatus = true;
  return true;
}

/*!
 *    @brief  Publishes the payload to the MQTT topic
 *    @param  topic
 *            the MQTT topic
 *    @param  payload
 *            the MQTT payload
 *    @return True if publish was successful.
 */
boolean GEENYmodem::mqttPublish(const char *topic, char *payload)
{
  debug->println("GEENYmodem::mqttPublish");
  if (!this->isMQTTConnected())
  {
    if (!this->connectMQTT())
    {
      return false;
    }
  }

  return (mqtt->publish(topic, payload));
}

/*!
 *    @brief  Connects to GPRS network.
 *    @return True if connection to GPRS network was successful.
 */
boolean GEENYmodem::connectGPRS(void)
{
  debug->println("GEENYmodem::connectGPRS");

  /*
    if(strlen(gprs->getIMSI().c_str())==0)
    {
      debug->println("GEENYmodem not installed");
      return false;
    }
    */

  if (!gprs->gprsConnect(APN, GPRS_USER, GPRS_PWD))
  {
    return false;
  }
  return true;
}

/*!
 *    @brief  Gets the MQTT credentials from tingg.io Modemmanager. This method needs to be called before any MQTT connection can be established
 *    @return True if connection to modemmanager powerup was successful.
 */
boolean GEENYmodem::powerup(void)
{
  debug->println("GEENYmodem::powerup");
  if (!connectGPRS())
  {
    debug->println("GPRS network not available");
    return false;
  }

  // close any connection before send a new request.
  // This will free the socket on the WiFi shield
  if (client->connected())
  {
    client->stop();
    delay(1000);
  }
  bool connected = false;

  debug->println(getIMEI().c_str());
  debug->println(getIMSI().c_str());
  debug->println(getOperator().c_str());
  debug->println(getLocalIP().c_str());

  // Make a HTTP POST request:
  debug->println(F("Performing HTTP POST request to /powerup..."));

  if (!client->connect(MODEM_MANAGER_URL, MODEM_MANAGER_PORT))
  {
    debug->println(F("can't connect to ModemManager"));
    return false;
  }
  else
  {
    debug->println(F("Connection successful, fetching MQTT credentials..."));

    //clear buffer
    memset(&buffer[0], 0, sizeof(buffer));

    // construct the HTTP POST request:
    sprintf_P(buffer, PSTR("POST %s HTTP/1.1\nHost: %s\nConnection: close\n"), MODEM_MANAGER_POWERUP_RESOURCE, MODEM_MANAGER_URL);
    debug->println(buffer);

    // send the HTTP POST request:
    client->print(buffer);

    // send empty line to end the request
    if (client->println() == 0)
    {
      debug->println(F("Failed to send request"));
    }

    uint16_t timeout = 0;
    // allow the response to be computed

    while (timeout <= 5000)
    {
      delay(10);
      timeout = timeout + 10;
      if (client->available())
      {
        break;
      }
    }

    // Check HTTP status
    char status[32] = {0};
    client->readBytesUntil('\r', status, sizeof(status));
    if (strcmp(status, "HTTP/1.1 200 OK") != 0)
    {
      debug->print(F("Unexpected response: "));
      debug->println(status);
      return false;
    }

    // Skip HTTP headers
    char endOfHeaders[] = "\r\n\r\n";
    if (!client->find(endOfHeaders))
    {
      debug->println(F("Invalid response"));
      return false;
    }

    // Allocate the JSON document
    // Use arduinojson.org/v6/assistant to compute the capacity.
    const size_t capacity = JSON_OBJECT_SIZE(2) + 93;

    DynamicJsonDocument doc(capacity);

    // Parse JSON object
    DeserializationError error = deserializeJson(doc, *client);
    if (error)
    {
      debug->print(F("deserializeJson() failed: "));
      debug->println(error.f_str());
      return false;
    }

    if (!doc["thing_id"])
    {
      debug->println(F("ERROR: thing_id is missing in JSON response from ModemManager!"));
      return false;
    }

    if (!doc["thing_key"])
    {
      debug->println(F("ERROR: thing_key is missing in JSON response from ModemManager!"));
      return false;
    }

    // Extract values
    debug->println(F("Response:"));
    debug->println(doc["thing_id"].as<char *>());
    debug->println(doc["thing_key"].as<char *>());

    //copy credentials to global variables
    strlcpy(this->thing_id, doc["thing_id"].as<char *>(), sizeof(this->thing_id));
    strlcpy(this->thing_key, doc["thing_key"].as<char *>(), sizeof(this->thing_key));

    // Disconnect
    client->stop();
    poweredup = true;
    return true;
  }
  return false;
}

/*!
 *    @brief  this method is only just during manufacturing for activating the modem in the tingg modem manager
 *    @param  x_api_key
 *            The API Key that was issued to you by tingg.io for waking up devices
 *    @return True if connection to modemmanager wakeup was successful.
 */
boolean GEENYmodem::wakeup(char *x_api_key)
{
  debug->println("GEENYmodem::wakeup");
  if (!connectGPRS())
  {
    debug->println("GPRS network not available");
    return false;
  }

  // close any connection before send a new request.
  // This will free the socket on the WiFi shield
  if (client->connected())
  {
    client->stop();
    delay(1000);
  }
  bool connected = false;

  // Make a HTTP POST request:
  debug->println(F("Performing HTTP POST request to /wakeup..."));

  if (!client->connect(MODEM_MANAGER_URL, MODEM_MANAGER_PORT))
  {
    debug->println(F("can't connect to ModemManager"));
    return false;
  }
  else
  {
    debug->println(F("Connection successful, activating (wakeup) device with modem manager..."));

    //clear buffer
    memset(&buffer[0], 0, sizeof(buffer));

    // construct the HTTP POST request:
    int contentLength = 21;
    contentLength += strlen(getIMSI().c_str());
    contentLength += strlen(getIMEI().c_str());

    sprintf_P(buffer, PSTR("POST %s HTTP/1.1\nHost: %s\nConnection: close\nx-api-key: %s\nContent-Type: application/json\nContent-Length: %d\r\n\r\n{\"imsi\":\"%s\",\"imei\":\"%s\"}"), MODEM_MANAGER_WAKEUP_RESOURCE, MODEM_MANAGER_URL, x_api_key, contentLength, getIMSI().c_str(), getIMEI().c_str());
    debug->println(buffer);

    // send the HTTP POST request:
    client->print(buffer);

    // send empty line to end the request
    if (client->println() == 0)
    {
      debug->println(F("Failed to send request"));
    }

    uint16_t timeout = 0;
    // allow the response to be computed

    while (timeout <= 5000)
    {
      delay(10);
      timeout = timeout + 10;
      if (client->available())
      {
        break;
      }
    }

    // Check HTTP status
    char status[32] = {0};
    client->readBytesUntil('\r', status, sizeof(status));
    if (strcmp(status, "HTTP/1.1 200 OK") != 0)
    {
      debug->print(F("Unexpected response: "));
      debug->println(status);
      return false;
    }

    // Skip HTTP headers
    char endOfHeaders[] = "\r\n\r\n";
    if (!client->find(endOfHeaders))
    {
      debug->println(F("Invalid response"));
      return false;
    }

    // Allocate the JSON document
    // Use arduinojson.org/v6/assistant to compute the capacity.
    const size_t capacity = JSON_OBJECT_SIZE(2) + 93;

    DynamicJsonDocument doc(capacity);

    // Parse JSON object
    DeserializationError error = deserializeJson(doc, *client);
    if (error)
    {
      debug->print(F("deserializeJson() failed: "));
      debug->println(error.f_str());
      return false;
    }

    if (!doc["code"])
    {
      debug->println(F("ERROR: code is missing in JSON response from ModemManager!"));
      return false;
    }

    // Extract values
    debug->println(F("CODE:"));
    debug->println(doc["code"].as<char *>());

    // Disconnect
    client->stop();
    return true;
  }
  return false;
}
